
/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 * Dancing Particules
 * Copyright (C) 2002 Pierre Tardy
 *
 * based on
 * Bezier Plugin
 * Copyright (C) 1999 Rick Haines
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <gtk/gtk.h>

#include <GL/gl.h>
#include <GL/glx.h>

#include <pthread.h>
#include <sched.h>


#include <math.h>
#include <unistd.h>
#include <stdio.h>

#include "xmms/xmmsctrl.h"
#include "xmms/configfile.h"
#include "xmms/plugin.h"
#include "gl.h"
#include "gl_plugin.h"
#include <stdlib.h>

#include "etoile.h"

/* These define the resolution and tonal scale of the spectrum analyzer */
#define NUM_BANDS       32

GLfloat heights[64];
pthread_mutex_t height_mutex;

static Display *display;
static Colormap colormap = 0;
static GLXContext glxcontext = NULL;
static Window root, window = 0;
static Atom wm_delete_window_atom;
static int x, y, w, h;

static Cursor no_ptr;
static Pixmap bm_no;
static unsigned char bm_no_data[] = { 0,0,0,0, 0,0,0,0 };

static pthread_t draw_thread;
static pthread_mutex_t still_going;

gboolean paused;

void bezier_init(void);
static void bezier_cleanup(void);
static void bezier_playback_start(void);
static void bezier_playback_stop(void);
static void bezier_render_freq(gint16 data[2][256]);
static void about_box();

static void init_mutexes(void);
static void destroy_mutexes(void);
static void create_window(void);
static void destroy_window(void);
static void *draw_thread_loop(void *arg);
static int detect_beat(gint32 loudness, gint *thickref, gint *quietref);

bezier_config bezier_cfg=
{
  0,0,0,0,1,1
};

VisPlugin bezier_vplugin = {
  NULL,                      /* internal    */
  NULL,                      /* internal    */
  0,                         /* internal    */
  NULL,                      /* description */
  0,                         /* # of pcm channels wanted    */
  256,                         /* # of freq channels wanted   */
  bezier_init,               /* called when plugin enabled  */
  bezier_cleanup,            /* called when plugin disabled */
  about_box,                      /* display about box           */
  NULL,//bezier_configure,   /* display configure box       */
  NULL,                      /* called to disable plugin    */
  bezier_playback_start,     /* called when playback starts */
  bezier_playback_stop,      /* called when playback stops  */
  NULL,                      /* called with pcm data        */
  bezier_render_freq         /* called with freq data       */
};

extern "C"
{
VisPlugin* get_vplugin_info(void)
{
#define VERSION "0.2"
  bezier_vplugin.description =
    g_strdup_printf("OpenGL dancing particules Plugin %s", VERSION);
  return &bezier_vplugin;
}
}
void bezier_read_config(void)
{
}

void bezier_write_config(void)
{

}

void init_vidmodes(void)
{
}

void bezier_init(void)
{
  bezier_read_config();
  init_mutexes();
  init_surface();
  pthread_mutex_lock(&still_going);
  pthread_create(&draw_thread, NULL, draw_thread_loop, NULL);
}

void bezier_cleanup(void)
{
  if (pthread_mutex_trylock(&still_going)) {
    pthread_mutex_unlock(&still_going);
    pthread_join(draw_thread, NULL);
  }
  destroy_mutexes();
}

void bezier_playback_start(void)
{
}

void bezier_playback_stop(void)
{
  int i;
  for(i = 0; i < 64; i++) 
    heights[i] = 0;
}
static gint xscale[] = {
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
        11, 12, 14, 16, 19, 22, 25, 28, 36, 42,
        49, 57, 66, 75, 85, 96, 107, 119, 132, 156,
        171, 187, 204, 220, 237, 255};  /* There are some extra bands here */

void bezier_render_freq(gint16 spectrum[2][256])
{
  gint    i,c;
  gint    y0, y1;
  gint32  loudness;


  pthread_mutex_lock(&height_mutex);

  for(i = 0; i < 32; i++) 
    heights[i] = (heights[i] + spectrum[0][i]) / 2;

  for(i = 0, loudness = 0; i < NUM_BANDS; i++)
    {
      /* find the average within each bar */
      for(c = xscale[i], y0 = y1 = 0; c < xscale[i + 1]; c++)
	{
	  y0 += spectrum[0][c];
	  y1 += spectrum[1][c];
	}
      y0 /= xscale[i+1] - xscale[i] + 1;
      y1 /= xscale[i+1] - xscale[i] + 1;

      /* Add this value to the loudness, scaling it so we are less
       * sensitive to midrange sounds.  Also, the spectrum data is
       * naturally less sensitive to high frequencies, so we need to
       * boost those.
       */
      //      loudness += y0 * (abs(i - NUM_BANDS/2) + NUM_BANDS/2) * (4 + i);
      //loudness += y0 * (((i-NUM_BANDS/2)*(i-NUM_BANDS/2)+60)/6);
      loudness += y0 * ((i-16)*(i-16)*(i-15)*(i-17))/600;
    }

  /* Scale loudness to be in about the same range as PCM loudness */
  loudness /= (NUM_BANDS * 4);

  int beat1 = detect_beat(loudness, &thick, &quiet);
  if(beat1)
    {
      beat=2;
      //      printf("bob!!\n");
    }
  if(beat1)
    beatcnt++;

  pthread_mutex_unlock(&height_mutex);
}

static gint disable_bezier_vplugin(gpointer data)
{
  bezier_vplugin.disable_plugin(&bezier_vplugin);
  return FALSE;
}

/* ***************************************************************** */

static void init_mutexes(void)
{
  pthread_mutex_init(&height_mutex, NULL);
  pthread_mutex_init(&still_going, NULL);
}

static void destroy_mutexes(void)
{
  pthread_mutex_destroy(&height_mutex);
  pthread_mutex_destroy(&still_going);
}

static void x_configure(XEvent *event)
{
  glViewport(0, 0, event->xconfigure.width, event->xconfigure.height);
}

static void x_message(Atom *atom)
{
  if (*atom == wm_delete_window_atom) {
    /* need to get this in the main thread */
    GDK_THREADS_ENTER();
    gtk_idle_add(disable_bezier_vplugin, NULL);
    GDK_THREADS_LEAVE();
  }
}

static void check_x_events(void)
{
  while (XPending(display)) {
    XEvent event;

    XNextEvent(display, &event);
    switch(event.type) {
    case ConfigureNotify:
      x_configure(&event);
      break;
    case ClientMessage:
      x_message((Atom*)&event.xclient.data.l[0]);
      break;
    }
  }
}

static void disable_fullscreen(void)
{

}

static void enable_fullscreen(void)
{
}

static void create_window(void)
{
  int attr_list[] = {
    GLX_RGBA,
    GLX_DEPTH_SIZE, 16,
    GLX_DOUBLEBUFFER,
    None
  };
  int screen_num;
  XSetWindowAttributes attr;
  unsigned long mask;
  XVisualInfo *vis_info;
  Atom wm_protocols[1];
  XColor black, dummy;

  display = XOpenDisplay(NULL);
  if (!display) return;

  screen_num = DefaultScreen(display);
  root = RootWindow(display, screen_num);

  vis_info = glXChooseVisual(display, screen_num, attr_list);
  if (!vis_info) return;

  attr.background_pixel = 0;
  attr.border_pixel = 0;
  attr.colormap = colormap = XCreateColormap(display, root, vis_info->visual, AllocNone);

  attr.event_mask = StructureNotifyMask | KeyPressMask;
  mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
  x = y = 0;
  w = 640; h = 480;

  window = XCreateWindow(display, root, x, y, w, h,
			 0, vis_info->depth, InputOutput,
			 vis_info->visual, mask, &attr);
  XmbSetWMProperties(display, window, "OpenGL Plugin",
		     "OpenGL Plugin", NULL, 0, NULL, NULL, NULL);
  wm_delete_window_atom = wm_protocols[0] = 
    XInternAtom(display, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(display, window, wm_protocols, 1);

  bm_no = XCreateBitmapFromData(display, window, (const char *)bm_no_data, 8, 8);
  XAllocNamedColor(display, colormap, "black", &black, &dummy);
  no_ptr = XCreatePixmapCursor(display, bm_no, bm_no, &black, &black, 0, 0);

  glxcontext = glXCreateContext(display, vis_info, NULL, True);

  glXMakeCurrent(display, window, glxcontext);

  XMapWindow(display, window);
  x = y = 50;
  XMoveWindow(display, window, x, y);
}

static void destroy_window(void)
{
  if (glxcontext) {
    glXMakeCurrent(display, 0, NULL);
    glXDestroyContext(display, glxcontext);
    glxcontext = NULL;
  }
  if (window) {
    XDestroyWindow(display, window);
    window = 0;
  }
  if (colormap) {
    XFreeColormap(display, colormap);
    colormap = 0;
  }
  if (display) {
    XCloseDisplay(display);
    display = NULL;
  }
}

static void set_priority(void)
{
#ifdef HAVE_SCHED_SETSCHEDULER
  struct sched_param sparam;
  sparam.sched_priority = sched_get_priority_max(SCHED_OTHER);
  pthread_setschedparam(pthread_self(), SCHED_OTHER, &sparam);		
#endif
}

static void *draw_thread_loop(void *arg)
{
  create_window();
  if (!display || !window) pthread_exit(NULL);
  init_vidmodes();
  init_gl();
  etoileinit();
  draw_gl();
  glXSwapBuffers(display, window);

  set_priority();
  paused = FALSE;
  while (pthread_mutex_trylock(&still_going)) {
    check_x_events();
    if (!paused) {
      draw_gl(); 
      glXSwapBuffers(display, window);
    }
#if 0
    sched_yield();
#endif
  }
  destroy_window();
  pthread_exit(NULL);
}

/* taken from blursk.c */

/*  Blursk - visualization plugin for XMMS
 *  Copyright (C) 1999  Steve Kirkendall
 *
 *  Portions of this file are derived from the XMMS "Blur Scope" plugin.
 *  XMMS is Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <stdlib.h>

/* To detect rhythms, blursk tracks the loudness of the signal across multiple
 * frames.  This is the maximum number of frames, and should correspond to the
 * slowest rhythm.
 */
#define BEAT_MAX	200
static gint32	beathistory[BEAT_MAX];
static int	beatbase;
static int	beatquiet;	/* force "quiet" situation? */
int gloudness;
/* Detect beats.  This involves more than just comparing the loudness to a
 * preset trigger level -- it also compensates for overall loudness of the
 * music, and even tries to detect the rhythm.
 *
 * It returns TRUE for beats, FALSE otherwise.  It also sets *thickref to a
 * thickness value from 0 to 3, and it detects the start of silence for
 * the "Random quiet" setting.
 */
static int detect_beat(gint32 loudness, gint *thickref, gint *quietref)
{
	static gint32	aged;		/* smoothed out loudness */
	static gint32	lowest;		/* quietest point in current beat */
	static int	elapsed;	/* frames since last beat */
	static int	isquiet;	/* was previous frame quiet */
	static int	prevbeat;	/* period of previous beat */
	int		beat, i, j;
	gint32		total;
	int		sensitivity;

	gloudness=loudness;
	/* Incorporate the current loudness into history */
	aged = (aged * 3 + loudness) >> 2;
	elapsed++;

	/* If silent, then clobber the beat */
	if (aged < 2000 || elapsed > BEAT_MAX)
	{
		elapsed = 0;
		lowest = aged;
		memset(beathistory, 0, sizeof beathistory);
	}
	else if (aged < lowest)
		lowest = aged;

	/* Beats are detected by looking for a sudden loudness after a lull.
	 * They are also limited to occur no more than once every 15 frames,
	 * so the beat flashes don't get too annoying.
	 */
	j = (beatbase + elapsed) % BEAT_MAX;
	beathistory[j] = loudness - aged;
	beat = FALSE;
	if (elapsed > 1 && aged > 2000 && loudness * 4 > aged * 5)
	{
		/* Compute the average loudness change, assuming this is beat */
		for (i = BEAT_MAX / elapsed, total = 0;
		     --i > 0;
		     j = (j + BEAT_MAX - elapsed) % BEAT_MAX)
		{
			total += beathistory[j];
		}
		total = total * elapsed / BEAT_MAX;

		/* Tweak the sensitivity to emphasize a consistent rhythm */
		sensitivity = 10;//config.beat_sensitivity;
		i = 3 - abs(elapsed - prevbeat)/2;
		if (i > 0)
		sensitivity += i;

		/* If average change is significantly positive, this is a beat.
		 */
		if (total * sensitivity > aged)
		{
			prevbeat = elapsed;
			beatbase = (beatbase + elapsed) % BEAT_MAX;
			lowest = aged;
			elapsed = 0;
			beat = TRUE;
		}
	}

	/* Thickness is computed from the difference between the instantaneous
	 * loudness and the aged loudness.  Thus, a sudden increase in volume
	 * will produce a thick line, regardless of rhythm.
	 */
#if 0
	if (aged < 1500)
		*thickref = 0;
	else if (!config.thick_on_beats)
		*thickref = 1;
	else
	{
		*thickref = loudness * 2 / aged;
		if (*thickref > 3)
			*thickref = 3;
	}
#endif
	/* Silence is computed from the aged loudness.  The quietref value is
	 * set to TRUE only at the start of silence, not throughout the silent
	 * period.  Also, there is some hysteresis so that silence followed
	 * by a slight noise and more silence won't count as two silent
	 * periods -- that sort of thing happens during many fade edits, so
	 * we have to account for it.
	 */
	if (beatquiet || aged < (isquiet ? 1500 : 500))
	{
		/* Quiet now -- is this the start of quiet? */
		*quietref = !isquiet;
		isquiet = TRUE;
		beatquiet = FALSE;
	}
	else
	{
		*quietref = FALSE;
		isquiet = FALSE;
	}

	/* return the result */
	return beat;
}

#define TITLEFONT "-*-helvetica-bold-r-normal--*-180-*-*-*-*-*-*"
#define TEXTFONT "-*-courier-medium-r-normal--*-120-*-*-*-*-*-*"

typedef struct
{
	gboolean  open;
	GtkWidget *window;
	GtkWidget *close;
	GtkWidget *text;
} dialog_t;
static dialog_t about_dialog;
static dialog_t error_dialog;
extern char readme[];
static void close_cb(GtkWidget *w, gpointer data)
{
	if (about_dialog.open &&
		(w == about_dialog.window || w == about_dialog.close))
	{
		gtk_widget_destroy(about_dialog.window);
		about_dialog.open = FALSE;
	}
	else if (error_dialog.open &&
		(w == error_dialog.window || w == error_dialog.close))
	{
		gtk_widget_destroy(error_dialog.window);
		error_dialog.open = FALSE;
	}
}

static void addtext(GtkWidget *twidget, const char *text)
{
	GdkFont *font;
	const char	*body;
	int	len;

	/* find the end of first line */
	for (len = 0, body = text; *body && *body != '\n'; body++, len++)
	{
	}
	if (*body == '\n');
		body++, len++;

	/* Add the first line in the "TITLEFONT" */
	font = gdk_font_load(TITLEFONT);
	gtk_text_insert(GTK_TEXT(twidget), font, NULL, NULL, text, len);
	gdk_font_unref(font);

	/* If there's any remaining text, add it in "TEXTFONT" */
	if (*body)
	{
		font = gdk_font_load(TEXTFONT);
		gtk_text_insert(GTK_TEXT(twidget), font, NULL, NULL, body, -1);
		gdk_font_unref(font);
	}
}

static dialog_t showtext(const char *text, const char *close_label)
{
	GtkWidget *vbox, *scroll, *bbox;
	dialog_t ret;

	/* Create a new window */
	ret.window = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_border_width(GTK_CONTAINER(ret.window), 10);
	gtk_window_position(GTK_WINDOW(ret.window), GTK_WIN_POS_MOUSE);

	/* Create a vbox inside the window */
	vbox = gtk_vbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(ret.window), vbox);

	/* Create a scrolled window in the vbox */
	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, FALSE, FALSE, 0);

	/* Create a text widget in the scrolled window */
	ret.text = gtk_text_new(NULL, NULL);
	gtk_widget_set_usize(ret.text, 450, 300);
	gtk_container_add(GTK_CONTAINER(scroll), ret.text);

	/* Stuff the text into it, using the Courier font */
	addtext(ret.text, text);
	gtk_text_set_editable(GTK_TEXT(ret.text), FALSE);

	/* Create a button box under the text widget */
	bbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 5);
	gtk_box_pack_start(GTK_BOX(vbox), bbox, FALSE, FALSE, 0);

	/* Add a [Close] button */
	ret.close = gtk_button_new_with_label(close_label);
	gtk_signal_connect_object(GTK_OBJECT(ret.close), "clicked", GTK_SIGNAL_FUNC(close_cb), GTK_OBJECT(ret.window));
	GTK_WIDGET_SET_FLAGS(ret.close, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(bbox), ret.close, TRUE, TRUE, 0);
	gtk_widget_grab_default(ret.close);

	/* Also make the window manager's "close" button close the window */
	gtk_signal_connect_object(GTK_OBJECT(ret.window), "delete_event", GTK_SIGNAL_FUNC(close_cb), GTK_OBJECT(ret.window));

	/* Show everything */
	gtk_widget_show(ret.close);
	gtk_widget_show(bbox);
	gtk_widget_show(ret.text);
	gtk_widget_show(scroll);
	gtk_widget_show(vbox);
	gtk_widget_show(ret.window);

	ret.open = TRUE;
	return ret;
}

void about_box()
{
	/* If already showing the window, do nothing */
	if (about_dialog.open)
		return;

	about_dialog = showtext(readme, "Close");
}

void about_error(char *format, ...)
{
	char	buf[2000];
	va_list	args;

	/* format the error message */
	va_start(args, format);
	vsprintf(buf, format, args);
	va_end(args);

	/* if no error window yet, then create one */
	if (error_dialog.open)
		addtext(error_dialog.text, buf);
	else
		error_dialog = showtext(buf, "Shit Happens");
}

/*main(int argc, char **argv)
{
	gtk_init(&argc, &argv);
	about();
	gtk_main();
	return 0;
}*/

